//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

package com.aleax.mufu.world2;

final class Vector2
{      
   double x,y;
   
   public Vector2(double x, double y)
   {
      this.x = x;
      this.y = y;
   }
   
   //(returns a formatted string containing x,y)
   public String toString()
   {
      return ( "(" + this.x + "," + this.y + ")" );
   }
   
   
   //----- these functions return Vector2s -----
   
   //return a copy of this
   public Object clone()
   {
      return new Vector2(this.x, this.y);
   }
   
   //return this+v2
   public Vector2 plus(Vector2 v2)
   {
      return new Vector2( this.x + v2.x, this.y + v2.y);
   }
   
   //return this-v2
   public Vector2 minus(Vector2 v2)
   {
      return new Vector2( this.x - v2.x, this.y - v2.y);
   }
   
   //return the righthand normal of this (using flash's coordinate system)
   public Vector2 normR()
   {
      return new Vector2( this.y * -1, this.x);
   }
   
   //return the (unit) direction vector of this
   public Vector2 dir()
   {
      Vector2 v = (Vector2)this.clone();
      v.normalize();
      return v;
   }
   
   //return this projected _onto_ v2
   public Vector2 proj(Vector2 v2)
   {
      double den = v2.dot(v2);
      
      if (den == 0)
      {
         //zero-length v2
         System.err.println("WARNING! Vector2.proj() was given a zero-length projection vector!");
         return (Vector2) this.clone();//not sure how to gracefully recover but, hopefully this will be okay
      }
      else
      {
         Vector2 v = (Vector2)v2.clone();
         v.mult(this.dot(v2)/den);
         return v;
      }
   }
   
   //return the magnitude (absval) of this projected onto v2
   public double projLen(Vector2 v2)
   {
      double den = v2.dot(v2);

      if( den == 0)
      {
         //zero-length v2
         System.err.println("WARNING! Vector2.projLen() was given a zero-length projection vector!");
         return 0;
      }
      else
      {
         return Math.abs(this.dot(v2)/den);
      }
   }
   
   
   //----- these functions return scalars -----
   
   //return the dotprod of this and v2
   public double dot(Vector2 v2)
   {
      return ((this.x * v2.x) + (this.y * v2.y));
   }
   
   //return the crossprod of this and v2
   //note that this is equivalent to the dotprod of this and the lefthand normal of v2
   public double cross(Vector2 v2)
   {
      return ((this.x * v2.y) - (this.y * v2.x) );
   }
   
   ///return the length of this
   public double len()
   {
      return ( Math.sqrt((this.x*this.x) + (this.y*this.y)) );
   }
   
   
   //----- these functions return nothing (they operate on this) -----
   
   //change this to a duplicate of v2
   public void copy(Vector2 v2)
   {
      this.x = v2.x;
      this.y = v2.y;
   }
   
   //multiply this by a scalar s
   public void mult(double s)
   {
      this.x *= s;
      this.y *= s;
   }
   
   //convert this vector to a unit/direction vector
   public void normalize()
   {
      double L = this.len();
      
      if( L != 0)
      {
         this.x /= L;
         this.y /= L;   
      }
      else
      {
         System.err.println("WARNING! Vector2.normalize() was called on a zero-length vector!");
      }
   }
   
   //add v2 to this
   public void pluseq(Vector2 v2)
   {
      this.x += v2.x;
      this.y += v2.y;
   }
   
   //subtract v2 from this
   public void minuseq(Vector2 v2)
   {
      this.x -= v2.x;
      this.y -= v2.y;
   }
}
