//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

package com.aleax.mufu.world2;

import java.awt.*;
import java.awt.geom.*;
import java.awt.image.BufferedImage;

/**
VectorRenderer provides simple rendering of lines/etc.
*/

final class VectorRenderer
{
   private final double TAN_SIXTEENTH = Math.tan(22.5 * Math.PI / 180.0);
   private final double COS_EIGTH = Math.cos(45.0 * Math.PI / 180.0);
   
   private int thickness;
   private int rgb;
   private int alpha;
   final Buffer buffer;
   
   public VectorRenderer()
   {
      //create an empty mc to draw into
      //this.buffer = gfx.createEmptySprite();
      this.buffer = new Buffer(demo5.XMAX, demo5.YMAX);
         
      //store a "global" line style
      this.thickness = 0;
      this.rgb = 0x000000;
      this.alpha = 100;
   }
   
   public Image getImage()
   {
      return buffer.img;
   }
   
   //---------- Draw functions ----------
   
   //clears the buffer
   public void clear()
   {
      this.buffer.clear();
      this.buffer.lineStyle( this.thickness, this.rgb, this.alpha );
   }
   
   //sets the drawing style
   public void setStyle(int thick, int rgb, int alpha)
   {
      this.buffer.lineStyle(thick, rgb, alpha);
   }
   
   public void startFill(int x, int y, int rgb, int alpha)
   {
      this.buffer.moveTo(x,y);
      this.buffer.beginFill(rgb, alpha);
   }
   
   public void stopFill()
   {
      this.buffer.endFill();
   }
   
   //draws a line from va to vb (vector2s)
   public void drawLine(Vector2 va, Vector2 vb)
   {
      this.buffer.moveTo( va.x, va.y );
      this.buffer.lineTo( vb.x, vb.y );
   }
   
   //draw a line specified by scalars
   public void drawLine_S(double x0, double y0, double x1, double y1)
   {
      this.buffer.moveTo( x0, y0 );
      this.buffer.lineTo( x1, y1 );
   }
   
   //draws a polyline given an array of verts vList
   public void drawLinestrip(Vector2[] vList)
   {
      this.buffer.moveTo( vList[0].x, vList[0].y );
      
      for(int i = 0; i < vList.length; i++)
         this.buffer.lineTo(vList[i].x, vList[i].y );
   }
   
   
   //draws a wireframe triangle va,vb,vc
   //NOTE: we could just use DrawLine but that's needlessly slow
   public void drawTri(Vector2 va, Vector2 vb, Vector2 vc) 
   {
      this.buffer.moveTo( va.x, va.y );
      this.buffer.lineTo( vb.x, vb.y );
      this.buffer.lineTo( vc.x, vc.y );
      this.buffer.lineTo( va.x, va.y );
   }
   
   public void drawTri_S(double vax, double vay, double vbx, double vby, double vcx, double vcy)   
   {
      this.buffer.moveTo( vax, vay );
      this.buffer.lineTo( vbx, vby );
      this.buffer.lineTo( vcx, vcy );
      this.buffer.lineTo( vax, vay );
   }
   
   //draws a wireframe box w/ CCW verts a,b,c,d
   public void drawQuad(Vector2 a, Vector2 b, Vector2 c, Vector2 d)
   {
      this.buffer.moveTo(a.x, a.y);
      this.buffer.lineTo(b.x, b.y);
      this.buffer.lineTo(c.x, c.y);
      this.buffer.lineTo(d.x, d.y);
      this.buffer.lineTo(a.x, a.y);
      
   }
   //scalar version
   public void drawQuad_S(double ax, double ay, double bx, double by, double cx, double cy, double dx, double dy)
   {
      this.buffer.moveTo(ax, ay);
      this.buffer.lineTo(bx, by);
      this.buffer.lineTo(cx, cy);
      this.buffer.lineTo(dx, dy);
      this.buffer.lineTo(ax, ay);   
   }  
      
   //draws a small plus at v
   public void drawPlus(Vector2 v)
   {
      this.buffer.moveTo( v.x-1, v.y);
      this.buffer.lineTo( v.x+1, v.y);
      this.buffer.moveTo( v.x, v.y-1);
      this.buffer.lineTo( v.x, v.y+1);
      
   }
   
   public void drawPlus_S(double vx, double vy)
   {
      this.buffer.moveTo( vx-1, vy);
      this.buffer.lineTo( vx+1, vy);
      this.buffer.moveTo( vx, vy-1);
      this.buffer.lineTo( vx, vy+1);         
   }
   
   //same as DrawPlus but with variable radius
   public void drawPlusR(Vector2 v, double r)
   {
      this.buffer.moveTo( v.x-r, v.y);
      this.buffer.lineTo( v.x+r, v.y);
      this.buffer.moveTo( v.x, v.y-r);
      this.buffer.lineTo( v.x, v.y+r);
      
   }
      
   //draws a small plus at v
   public void drawCross(Vector2 v)
   {
      this.buffer.moveTo( v.x-1, v.y-1);
      this.buffer.lineTo( v.x+1, v.y+1);
      this.buffer.moveTo( v.x+1, v.y-1);
      this.buffer.lineTo( v.x-1, v.y+1);
      
   }  
   
   public void drawCross_S(double vx, double vy)
   {  
      this.buffer.moveTo( vx-1, vy-1);
      this.buffer.lineTo( vx+1, vy+1);
      this.buffer.moveTo( vx+1, vy-1);
      this.buffer.lineTo( vx-1, vy+1);
      
   }
   
   //same as DrawCross but with variable radius
   public void drawCrossR(Vector2 v, double r)
   {
      this.buffer.moveTo( v.x-r, v.y-r);
      this.buffer.lineTo( v.x+r, v.y+r);
      this.buffer.moveTo( v.x+r, v.y-r);
      this.buffer.lineTo( v.x-r, v.y+r);
      
   }  
      
   //this draws a circle at the point specified by v, with radius r
   // (code by casper s)
   public void drawCircle(Vector2 v, double r)
   {
      double x = v.x;
      double y = v.y;
   
      this.buffer.moveTo(x+r, y);
      this.buffer.curveTo(r+x, TAN_SIXTEENTH*r+y, COS_EIGTH*r+x, COS_EIGTH*r+y);
      this.buffer.curveTo(TAN_SIXTEENTH*r+x, r+y, x, r+y);
      this.buffer.curveTo(-TAN_SIXTEENTH*r+x, r+y, -COS_EIGTH*r+x, COS_EIGTH*r+y);
      this.buffer.curveTo(-r+x, TAN_SIXTEENTH*r+y, -r+x, y);
      this.buffer.curveTo(-r+x, -TAN_SIXTEENTH*r+y, -COS_EIGTH*r+x, -COS_EIGTH*r+y);
      this.buffer.curveTo(-TAN_SIXTEENTH*r+x, -r+y, x, -r+y);
      this.buffer.curveTo(TAN_SIXTEENTH*r+x, -r+y, COS_EIGTH*r+x, -COS_EIGTH*r+y);
      this.buffer.curveTo(r+x, -TAN_SIXTEENTH*r+y, r+x, y);
   }
   
   public void drawCircle_S(double x, double y, double r)
   {
      this.buffer.moveTo(x+r, y);
      this.buffer.curveTo(r+x, TAN_SIXTEENTH*r+y, COS_EIGTH*r+x, COS_EIGTH*r+y);
      this.buffer.curveTo(TAN_SIXTEENTH*r+x, r+y, x, r+y);
      this.buffer.curveTo(-TAN_SIXTEENTH*r+x, r+y, -COS_EIGTH*r+x, COS_EIGTH*r+y);
      this.buffer.curveTo(-r+x, TAN_SIXTEENTH*r+y, -r+x, y);
      this.buffer.curveTo(-r+x, -TAN_SIXTEENTH*r+y, -COS_EIGTH*r+x, -COS_EIGTH*r+y);
      this.buffer.curveTo(-TAN_SIXTEENTH*r+x, -r+y, x, -r+y);
      this.buffer.curveTo(TAN_SIXTEENTH*r+x, -r+y, COS_EIGTH*r+x, -COS_EIGTH*r+y);
      this.buffer.curveTo(r+x, -TAN_SIXTEENTH*r+y, r+x, y);
   }  
      
      
   //this draws an arc from p0 to p1, with control point c
   public void drawArc(Point p0, Point p1, Point c)
   {
      this.buffer.moveTo(p0.x,p0.y);
      this.buffer.curveTo(c.x, c.y, p1.x, p1.y);
   }
   
   //scalar version
   public void drawArc_S(double x0, double y0, double x1, double y1, double xc, double yc)
   {
      this.buffer.moveTo(x0,y0);
      this.buffer.curveTo(xc, yc, x1, y1);
   }
   
   //this draws an AABB centered at p, with halfwidths xw,yw
   public void drawAABB(Vector2 p, double xw, double yw)
   {
      Vector2 v0 = new Vector2(p.x+xw,p.y+yw);  
      Vector2 v1 = new Vector2(p.x-xw,p.y+yw);
      Vector2 v2 = new Vector2(p.x-xw,p.y-yw);
      Vector2 v3 = new Vector2(p.x+xw,p.y-yw);
      this.drawQuad(v0,v1,v2,v3);         
   }
   
   public void drawAABB_S(double minx, double maxx, double miny, double maxy)
   {
      //scalar version
      Vector2 v0 = new Vector2(maxx,maxy);   
      Vector2 v1 = new Vector2(minx,maxy);
      Vector2 v2 = new Vector2(minx,miny);
      Vector2 v3 = new Vector2(maxx,miny);
      this.drawQuad(v0,v1,v2,v3);   
   }
   
   //draws a 90deg circle arc _CCW_.
   //cx,cy is the circle center, px,py is the point on the circle surface at which to start drawing
   public void drawConcaveCCWArc_S(double cx, double cy, double px, double py)
   {
      double p0x = px;//first bezier point
      double p0y = py;
      
      double vx = p0x - cx;//radius vector
      double vy = p0y - cy;
      
      double r = Math.sqrt(vx*vx + vy*vy);//circle radius
      double nx = vy;//nx,ny is the lefthand normal of the radius vector
      double ny = -vx;
      
      double p1x = (p0x + nx) - cx;
      double p1y = (p0y + ny) - cy;
      double len = Math.sqrt(p1x*p1x + p1y*p1y);
      p1x /= len;
      p1y /= len;
      p1x *= r;
      p1y *= r;
      p1x += cx;//second bezier point
      p1y += cy;
      
      double c0x = ((p0x + p1x)*0.5) - cx;//this point is "inside" the circle; we want it to
      double c0y = ((p0y + p1y)*0.5) - cy;//be outside the circle by the same distance it's currently inside
      double clen = Math.sqrt(c0x*c0x + c0y*c0y);
      double dlen = r - clen;
      //dlen *= 2;//moving the control point along the radius vector by dlen will put it on the surface; we want it outside so we double the amount we displace it
      c0x /= clen;
      c0y /= clen;
      c0x *= (r+dlen);
      c0y *= (r+dlen);//c0x,c0y is now the control point between bezier points 0 and 1
      c0x += cx;
      c0y += cy;
      
      this.buffer.moveTo(p0x,p0y);
      this.buffer.curveTo(c0x,c0y,p1x,p1y);
         
      //now we repeat the process for the next 45degrees
      p0x = p1x;//first bezier point
      p0y = p1y;
      
      vx = p0x - cx;//radius vector
      vy = p0y - cy;
      
      r = Math.sqrt(vx*vx + vy*vy);
      nx = vy;//nx,ny is the lefthand normal of the radius vector
      ny = -vx;
      
      p1x = (p0x + nx) - cx;
      p1y = (p0y + ny) - cy;
      len = Math.sqrt(p1x*p1x + p1y*p1y);
      p1x /= len;
      p1y /= len;
      p1x *= r;
      p1y *= r;
      p1x += cx;//second bezier point
      p1y += cy;
      
      c0x = ((p0x + p1x)*0.5) - cx;//this point is "inside" the circle; we want it to
      c0y = ((p0y + p1y)*0.5) - cy;//be outside the circle by the same distance it's currently inside
      clen = Math.sqrt(c0x*c0x + c0y*c0y);
      dlen = r - clen;
      //dlen *= 2;//moving the control point along the radius vector by dlen will put it on the surface; we want it outside so we double the amount we displace it
      c0x /= clen;
      c0y /= clen;
      c0x *= (r+dlen);
      c0y *= (r+dlen);//c0x,c0y is now the control point between bezier points 0 and 1
      c0x += cx;
      c0y += cy;
      
      this.buffer.curveTo(c0x,c0y,p1x,p1y);
   }
   
   //----------------------------------------------------------------------------------------------
   
   static final class Buffer
   {
      final Image img;
      private final Graphics2D g;
      private final Point p = new Point(0,0);
      private GeneralPath shape;
      private Color fillColor;
      Color backgroundColor = Color.white;
      
      Buffer(int width, int height)
      {
         img = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
         g = (Graphics2D)img.getGraphics();
         
         g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);         
      }
      
      void moveTo(double x, double y)
      {
         if (shape!=null)
         {
            shape.moveTo((int)x, (int)y);
         }
         else
         {
            p.x = (int)x;
            p.y = (int)y;
         }
      }

      void lineTo(double x, double y)
      {
         if (shape!=null)
         {
            shape.lineTo((int)x,(int)y);
         }
         else
         {
            g.drawLine(p.x, p.y, (int)x, (int)y);
            moveTo(x,y);
         }
      }
      
      void lineStyle(int width, int rgb, int alpha)
      {
         alpha = alpha*0xff/100;
         g.setStroke(new BasicStroke(1.0f,BasicStroke.CAP_SQUARE,BasicStroke.JOIN_MITER));
         g.setColor(new Color(rgb|alpha<<24));
      }
      
      void clear()
      {
         g.setBackground(backgroundColor);
         g.clearRect(0,0,img.getWidth(null), img.getHeight(null));
      }
      
      void curveTo(double controlX, double controlY, double anchorX, double anchorY)
      {
         if (shape!=null)
         {
            shape.quadTo((int)controlX, (int)controlY, (int)anchorX, (int)anchorY);
         }
         else
         {
            g.draw(new QuadCurve2D.Double(p.x, p.y, controlX, controlY, anchorX, anchorY));
            moveTo(anchorX, anchorY);
         }
      }
      
      void beginFill(int rgb, int alpha)
      {
         shape = new GeneralPath();
         alpha = alpha*0xff/100;
         fillColor = new Color(rgb|alpha<<24);
      }
      
      void endFill()
      {
         Color oldColor = g.getColor();
         g.setColor(fillColor);
         g.fill(shape);
         g.setColor(oldColor);
         g.draw(shape);
         shape = null;
      }
   }
}     
