//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

package com.aleax.mufu.world2;

import java.awt.*;
import java.awt.event.*;
import java.util.HashSet;

import javax.swing.*;
import javax.swing.event.*;

/** 
 * Demo harness - provides convenience methods to make writing demos easier
 */

public class demo implements KeyListener, MouseListener, MouseMotionListener
{
   //app constants
   public static final int XMIN = 0;//these define the world bounds
   public static final int XMAX = 400;
   public static final int YMIN = 0;
   public static final int YMAX = 400;
   
   // Rendering
   private static final int TARGET_FPS = 60;
   private static long lastFrameTime;
   protected JPanel demoPanel;
   protected DynamicObject demoObj;
 
   // Entry point from OS
   public static void main(String[] args) throws Exception
   {
      String[] demoNames =  new String[]{"Demo 4", "Demo 5"};
      Class[] demoClasses =  new Class[]{demo4.class, demo5.class};
      
      int choice = JOptionPane.showOptionDialog(null, "Which demo to run?", "Select demo", 0, 0, null, demoNames, "");
      
      demoClasses[choice].newInstance();
   }
   
   /** Creates a panel with sliders to alter some of the demo object's properties */
   protected JPanel createSliders()   
   {
      final JSlider gSlider = new JSlider(0,150, (int)(demoObj.gravity*150));
      final JSlider bSlider = new JSlider(0,150, (int)(demoObj.bounce*150));
      final JSlider fSlider = new JSlider(0,150, 40);

      gSlider.addChangeListener
      (
         new ChangeListener() { public void stateChanged(ChangeEvent e){ demoObj.gravity = gSlider.getValue()/150.0 * DynamicObject.MAX_GRAVITY; } }
      );
      
      bSlider.addChangeListener
      (
         new ChangeListener() { public void stateChanged(ChangeEvent e){ demoObj.bounce = bSlider.getValue()/150.0 * DynamicObject.MAX_BOUNCE; } }
      );
      
      fSlider.addChangeListener
      (
         new ChangeListener() { public void stateChanged(ChangeEvent e){ demoObj.friction = Math.pow(fSlider.getValue()/150.0,3) * DynamicObject.MAX_FRICTION; } }
      );
      
      JPanel panel = new JPanel(new FlowLayout());
      Box box1 = Box.createVerticalBox();
      box1.add(new JLabel("Gravity:"));
      box1.add(new JLabel("Bounce:"));
      box1.add(new JLabel("Friction:"));
      Box box2 = Box.createVerticalBox();
      box2.add(gSlider);
      box2.add(bSlider);
      box2.add(fSlider);
      panel.add(box1);
      panel.add(box2);

      return panel;
   }
   
   /** Starts a thread to run the demo. Calls tickDemo() once per loop to update game world */
   public void runDemo()
   {
      //Kick off simulation thread
      new Thread()
      {
         public void run()
         {
            for (;;)
            {
               tickDemo();
               demoPanel.repaint();
               
               int sleepTime = 1000/TARGET_FPS;
               sleepTime = Math.min(sleepTime, (int)(lastFrameTime+1000/TARGET_FPS-System.currentTimeMillis()));

               if (sleepTime>0)
               {
                  try { Thread.sleep(sleepTime); }
                  catch (InterruptedException e){}
               }

               lastFrameTime = System.currentTimeMillis();
            }
         }
      }.start();      
   }
   
   /** Sub-classes should override this method to run their simulation */
   protected void tickDemo(){}

   //----------------------------------------------------------------------------------------------
   // methods to respond to user input
   
   protected final HashSet<Integer> heldKeys = new HashSet<Integer>();
   protected boolean mouseDown;
   protected final Point m = new Point();
      
   protected void addKeyListeners(Container c)
   {
      for (int i=0 ; i<c.getComponentCount() ; i++)
      {
         Component child = c.getComponent(i);
         child.addKeyListener(this);

         //Remove mappings for keys used in the demo UI
         if (child instanceof JComponent)
         {
            ((JComponent)child).getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_UP, 0), "none");
            ((JComponent)child).getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_DOWN, 0), "none");
            ((JComponent)child).getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_LEFT, 0), "none");
            ((JComponent)child).getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_RIGHT, 0), "none");
            ((JComponent)child).getInputMap().put(KeyStroke.getKeyStroke(KeyEvent.VK_SPACE, 0), "none");
         }
         
         if (child instanceof Container)
            addKeyListeners((Container)child);
      }
   }
   
   public void keyPressed(KeyEvent e)        { heldKeys.add(e.getKeyCode()); }
   public void keyReleased(KeyEvent e)       { heldKeys.remove(e.getKeyCode()); }
   public void keyTyped(KeyEvent e)          {}
   
   public void mouseMoved(MouseEvent me)     { m.setLocation(me.getPoint()); }
   public void mouseDragged(MouseEvent me)   { m.setLocation(me.getPoint()); }
   
   public void mouseClicked(MouseEvent e){}
   public void mouseEntered(MouseEvent e){}
   public void mouseExited(MouseEvent e){}
   public void mousePressed(MouseEvent e)    { mouseDown = true; }
   public void mouseReleased(MouseEvent e)   { mouseDown = false; }
}