//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

package com.aleax.mufu.world2;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

final class demo4 extends demo
{
   //app constants
   public static final int OBJRAD = 16;
   public static final int TILERAD = 20;
   public static final double OBJSPEED = 0.2;
   public static final double MAXSPEED = 20;

   // demo level
   private static final String MAPSTR = "A6E00002000?E000000NA0070C0N00;10B0N00:10>0>L0060000F000@0GH0003";
         
   // demo objects   
   private final TileMap tiles = new TileMap(8,8,TILERAD,TILERAD);//map is 10x10 tiles, minus a 1-tile border on each edge.
   private Vector2 RAY_O = new Vector2(0,0);
   
   // Rendering
   private VectorRenderer debug_rend = new VectorRenderer();
   private VectorRenderer ray_rend = new VectorRenderer();
   
   private demo4()   
   {
      demoObj = new Circle(new Vector2((XMIN+XMAX)/2, YMAX - 2*4*TILERAD), OBJRAD);
      
      System.out.println("building tilemap");      
      while (tiles.building()){}
         
      // begin the loading process
      tiles.setMap(MAPSTR);
      
      //Set background of debug renderer to transparent
      debug_rend.buffer.backgroundColor = new Color(0,0,0,0);
      ray_rend.buffer.backgroundColor = new Color(0,0,0,0);
      
      //Create the panel to display the demo
      final JPanel demoPanel = new JPanel()
      {
         public void paint(Graphics g) 
         {
            g.drawImage(tiles.rend.getImage(),0,0,null);
            g.drawImage(debug_rend.getImage(),0,0,null);
            g.drawImage(ray_rend.getImage(),0,0,null);
         }
      };
      
      demoPanel.setMinimumSize(new Dimension(XMAX-XMIN, YMAX-YMIN));
      demoPanel.setPreferredSize(demoPanel.getMinimumSize());
      
      //Create a sub-panel for the controls
      JButton changeObjButton = new JButton("change obj shape"); 
    
      changeObjButton.addActionListener
      (
         new ActionListener() { public void actionPerformed(ActionEvent ae) { toggleObjShape(); } }
      );

      JPanel controlPanel = new JPanel();
      controlPanel.add(changeObjButton, BorderLayout.WEST);
      controlPanel.add(createSliders(), BorderLayout.EAST);
      
      //Create a frame to hold the panels
      JFrame frame = new JFrame("Demo 5");      
      frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      frame.getContentPane().add(demoPanel, BorderLayout.CENTER);
      frame.getContentPane().add(controlPanel, BorderLayout.SOUTH);
      frame.pack();
      frame.setLocationByPlatform(true);
      frame.setResizable(false);
      frame.setVisible(true);

      addKeyListeners(frame);      
      demoPanel.addMouseListener(this);      
      demoPanel.addMouseMotionListener(this);

      runDemo();
   }

   protected void toggleObjShape()
   {
      Vector2 p = demoObj.pos;
      Vector2 o = demoObj.oldPos;
      
      if(demoObj instanceof Circle)
         demoObj = new AABB(new Vector2(p.x, p.y), OBJRAD*0.7, OBJRAD);
      else
         demoObj = new Circle(new Vector2(p.x, p.y), OBJRAD);
      
      demoObj.oldPos.x = o.x;
      demoObj.oldPos.y = o.y;
   }   
   
   protected void tickDemo()
   {
   	debug_rend.clear();
   	ray_rend.clear();
   
   	//only fire a ray if the mouse is inside the world
   	if (mouseDown && m.y < 400)
   	{
			//fire a beam with the mouse
			Vector2 result = new Vector2(0,0);

         double dx = m.x - RAY_O.x;
         double dy = m.y - RAY_O.y;
         double len = Math.sqrt(dx*dx + dy*dy);
         
			if(len != 0)
			{
				//the ray isn't degenerate; use it
	
				if (tiles.queryRayObj(result, RAY_O, new Vector2(m.x, m.y), demoObj))
				{
					//ray hit the object; shove the object a bit
					dx /= len;//get ray direction
					dy /= len;
						
               double rayshove = 1;//shove strength
						
					demoObj.pos.x += dx*rayshove;//shove the object
					demoObj.pos.y += dy*rayshove;					
				}
				
				//draw the ray; note that the ray should always hit the world, since the world is bordered.
				ray_rend.setStyle(0, 0x882288, 100);//draw ray
				ray_rend.drawLine(RAY_O, result);
						
				ray_rend.drawPlus(RAY_O);//draw origin
						
				ray_rend.setStyle(2, 0x882288, 100);//draw end
				ray_rend.drawCross(result);
			}
   	}
      		
		//update dynamic object
		demoObj.verlet();//integrate dynamic object's position forward in time
			
		//simple user interaction
		if(heldKeys.contains(KeyEvent.VK_SPACE))
		{		
			//let user apply force by dragging the mouse
         double dragScale = 0.1;
         demoObj.pos.x += (m.x-demoObj.pos.x)*dragScale;
         demoObj.pos.y += (m.y-demoObj.pos.y)*dragScale;
		}
		else
		{
			//thrust via arrow keys
         double fx = 0;
         double fy = 0;
         
			if(heldKeys.contains(KeyEvent.VK_LEFT))
				fx -= OBJSPEED;
			
         if(heldKeys.contains(KeyEvent.VK_RIGHT))
				fx += OBJSPEED;
			
         if(heldKeys.contains(KeyEvent.VK_UP))
				fy -= OBJSPEED + demoObj.gravity;

         if(heldKeys.contains(KeyEvent.VK_DOWN))
				fy += OBJSPEED;

         Vector2 p = demoObj.pos;
         Vector2 o = demoObj.oldPos;
         double vx = p.x - o.x;
         double vy = p.y - o.y;
         double newx = Math.min(MAXSPEED, Math.max(-MAXSPEED, vx+fx));
         double newy = Math.min(MAXSPEED, Math.max(-MAXSPEED, vy+fy));
			p.x = o.x + newx;
			p.y = o.y + newy;
		}
					
      demoObj.collideVsWorldBounds();
		demoObj.collideVsTileMap(tiles);//collide vs. tilemap   	
   			
   	//draw dynamic object.. hacky!
   	debug_rend.setStyle(0,0x797988,100);
   	debug_rend.buffer.beginFill(0xCACAD0,100);
   	demoObj.draw(debug_rend);
   	debug_rend.buffer.endFill();	   
   }
   
   //----------------------------------------------------------------------------------------------

   public void mousePressed(MouseEvent e)
   {
      //click; set ray origin       
      RAY_O.x = e.getX();
      RAY_O.y = e.getY();
      
      super.mousePressed(e);
   }
}