//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

package com.aleax.mufu.world2;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

final class demo5 extends demo
{
   //app constants
   public static final int OBJRAD = 24;
   public static final int TILERAD = OBJRAD*2;
   public static final double OBJSPEED = 0.2;
   public static final double MAXSPEED = 20;
   
   //instance data
   private int curTileShape = 0;
   private static final int[] tileShapeList = 
   {
      TileMapCell.TID_FULL, TileMapCell.TID_45DEGpn, TileMapCell.TID_CONCAVEnn, TileMapCell.TID_CONVEXpn, TileMapCell.TID_22DEGnnS, 
      TileMapCell.TID_22DEGpnB, TileMapCell.TID_67DEGnnS, TileMapCell.TID_67DEGpnB, TileMapCell.TID_HALFl
   };

   //Rendering
   private final VectorRenderer debug_rend = new VectorRenderer();
   private final VectorRenderer static_rend = new VectorRenderer();
   
   //Demo objects
   private TileMapCell demoTile;
   
   demo5()
   {
      demoObj = new Circle(new Vector2(0.5*(XMIN+XMAX), YMAX - 3*OBJRAD), OBJRAD);
      
      //draw the world bounds
      static_rend.clear();
      static_rend.setStyle(0,0x000000,100);
      static_rend.drawAABB_S(XMIN,XMAX-1,YMIN,YMAX-1);
      
      //Set background of debug renderer to transparent
      debug_rend.buffer.backgroundColor = new Color(0,0,0,0);
            
      //make a tile
      demoTile = new TileMapCell(0.5*(XMAX+XMIN),0.5*(YMAX+YMIN), TILERAD, TILERAD);
      demoTile.setState(TileMapCell.TID_FULL);
      
      //Create the panel to display the demo
      demoPanel = new JPanel()
      {
         public void paint(Graphics g) 
         {
            g.drawImage(static_rend.getImage(),0,0,null);
            g.drawImage(debug_rend.getImage(),0,0,null);
         }
      };
      
      demoPanel.setMinimumSize(new Dimension(XMAX-XMIN, YMAX-YMIN));
      demoPanel.setPreferredSize(demoPanel.getMinimumSize());
      
      //Create a sub-panel for the controls
      JButton changeObjButton = new JButton("change obj shape"); 
      JButton changeTileButton = new JButton("change tile shape");
      
      changeObjButton.addActionListener
      (
         new ActionListener() { public void actionPerformed(ActionEvent ae) { toggleObjShape(); } }
      );

      changeTileButton.addActionListener
      (
         new ActionListener() { public void actionPerformed(ActionEvent ae) { toggleTileShape(); } }
      );
      
      JPanel leftPanel = new JPanel(new BorderLayout(0,2));
      leftPanel.add(changeObjButton, BorderLayout.NORTH);
      leftPanel.add(changeTileButton, BorderLayout.SOUTH);

      JPanel controlPanel = new JPanel();
      controlPanel.add(leftPanel, BorderLayout.WEST);
      controlPanel.add(createSliders(), BorderLayout.EAST);
      
      //Create a frame to hold the panels
      JFrame frame = new JFrame("Demo 5");            
      frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      frame.getContentPane().add(demoPanel, BorderLayout.CENTER);
      frame.getContentPane().add(controlPanel, BorderLayout.SOUTH);
      frame.pack();
      frame.setLocationByPlatform(true);
      frame.setResizable(false);
      frame.setVisible(true);

      addKeyListeners(frame);      
      demoPanel.addMouseListener(this);      
      demoPanel.addMouseMotionListener(this);
      
      runDemo();
   }
   
   /** called from main loop */
   protected void tickDemo()
   {
   	debug_rend.clear();
   
		// update dynamic object   	
		
   	// integrate dynamic object's position forward in time
		demoObj.verlet();
			
		// simple user interaction
		if (mouseDown)
		{		
			//let user apply force by dragging the mouse
			double dragScale = 0.1;
			demoObj.pos.x += (m.x-demoObj.pos.x)*dragScale;
			demoObj.pos.y += (m.y-demoObj.pos.y)*dragScale;
		}
		else if(heldKeys.contains(KeyEvent.VK_SPACE))
		{
			// grab object with mouse
			demoObj.pos.x = m.x;
			demoObj.pos.y = m.y;		
		}
		else
		{
			// thrust via arrow keys
			double fx = 0;
			double fy = 0;
         
			if (heldKeys.contains(KeyEvent.VK_LEFT))
				fx -= OBJSPEED;

			if (heldKeys.contains(KeyEvent.VK_RIGHT))
				fx += OBJSPEED;
	
			if (heldKeys.contains(KeyEvent.VK_UP))
				fy -= OBJSPEED + demoObj.gravity;

			if (heldKeys.contains(KeyEvent.VK_DOWN))
				fy += OBJSPEED;
         
         Vector2 p = demoObj.pos;
         Vector2 o = demoObj.oldPos;
         double vx = p.x - o.x;
         double vy = p.y - o.y;
         double newx = Math.min(MAXSPEED, Math.max(-MAXSPEED, vx+fx));
         double newy = Math.min(MAXSPEED, Math.max(-MAXSPEED, vy+fy));
			p.x = o.x + newx;
			p.y = o.y + newy;
		}
				
		//collide dynamic object	
		demoObj.collideVsTile(demoTile);
		demoObj.collideVsWorldBounds();
   			
   	//draw dynamic object and tile.. hacky!
   	debug_rend.setStyle(0,0x797988,100);
   	debug_rend.buffer.beginFill(0xCACAD0,100);
   	demoObj.draw(debug_rend);
   	debug_rend.buffer.endFill();
      
      debug_rend.setStyle(0,0x797988,100);
      debug_rend.buffer.beginFill(0xCACAD0,100);
      demoTile.draw(debug_rend);
      debug_rend.buffer.endFill();      
   }
   
   //----------------------------------------------------------------------------------------------   
   // methods to toggle between object shapes (for both dynamic and static shapes)
   
   private void toggleObjShape()
   {
   	Vector2 p = demoObj.pos;
      Vector2 o = demoObj.oldPos;
      
   	if (demoObj instanceof Circle)
   		demoObj = new AABB(new Vector2(p.x, p.y), OBJRAD*0.75, OBJRAD*1.25);
   	else
   		demoObj = new Circle(new Vector2(p.x, p.y), OBJRAD);

   	demoObj.oldPos.x = o.x;
   	demoObj.oldPos.y = o.y;
   }
   
   private void toggleTileShape()
   {
   	curTileShape = (curTileShape+1)%tileShapeList.length;
   	demoTile.setState(tileShapeList[curTileShape]);	
   }
   
   //----------------------------------------------------------------------------------------------
   
   public void mouseMoved(MouseEvent me)
   {
      if ((me.getModifiers()&MouseEvent.SHIFT_MASK) != 0)
      {
         //grab tile with mouse; tiles are supposed to be static so this is a bit hacky
         demoTile.pos.x = Math.min(XMAX-TILERAD, Math.max(XMIN+TILERAD, me.getX()));
         demoTile.pos.y = Math.min(YMAX-TILERAD, Math.max(YMIN+TILERAD, me.getY()));
      }
      else
         super.mouseMoved(me);
   }
}